/*
 * Copyright (C) 2026 Samuel Weinig <sam@webkit.org>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#if ENABLE(THREADED_ANIMATIONS)

#include <WebCore/AcceleratedEffectCircleFunction.h>
#include <WebCore/AcceleratedEffectEllipseFunction.h>
#include <WebCore/AcceleratedEffectInsetFunction.h>
#include <WebCore/AcceleratedEffectPathFunction.h>
#include <WebCore/AcceleratedEffectPolygonFunction.h>
#include <WebCore/AcceleratedEffectShapeFunction.h>
#include <wtf/Variant.h>

namespace WebCore {

struct BlendingContext;
template<typename> struct AcceleratedEffectInterpolation;

struct AcceleratedEffectBasicShape {
    using Function = Variant<
        AcceleratedEffectCircleFunction,
        AcceleratedEffectEllipseFunction,
        AcceleratedEffectInsetFunction,
        AcceleratedEffectPathFunction,
        AcceleratedEffectPolygonFunction,
        AcceleratedEffectShapeFunction
    >;

    Function function;

    bool operator==(const AcceleratedEffectBasicShape&) const = default;
};

template<typename T> concept AcceleratedEffectShapeWithCenterCoordinate = std::same_as<T, AcceleratedEffectCircleFunction> || std::same_as<T, AcceleratedEffectEllipseFunction>;

// MARK: - Blending

template<> struct AcceleratedEffectInterpolation<AcceleratedEffectBasicShape> {
    auto canBlend(const AcceleratedEffectBasicShape&, const AcceleratedEffectBasicShape&) -> bool;
    auto blend(const AcceleratedEffectBasicShape&, const AcceleratedEffectBasicShape&, const BlendingContext&) -> AcceleratedEffectBasicShape;
};

} // namespace WebCore

#endif // ENABLE(THREADED_ANIMATIONS)
