/*
 * Copyright (C) 2013 Google Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above
 * copyright notice, this list of conditions and the following disclaimer
 * in the documentation and/or other materials provided with the
 * distribution.
 *     * Neither the name of Google Inc. nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
 
enum EndOfStreamError {
    "network",
    "decode"
};

enum ReadyState {
    "closed",
    "open",
    "ended"
};

[
    JSGenerateToNativeObject,
    ActiveDOMObject,
    Conditional=MEDIA_SOURCE,
    ConditionalForWorker=MEDIA_SOURCE_IN_WORKERS,
    EnabledBySetting=MediaSourceEnabled,
    EnabledForContext,
    Exposed=(Window,DedicatedWorker)
] interface MediaSource : EventTarget {
    [CallWith=CurrentScriptExecutionContext] constructor(optional MediaSourceInit init = {});

    [EnabledBySetting=DetachableMediaSourceEnabled] readonly attribute boolean detachable;

    [SameObject, Exposed=DedicatedWorker, Conditional=MEDIA_SOURCE_IN_WORKERS, EnabledBySetting=MediaSourceInWorkerEnabled]
    readonly attribute MediaSourceHandle handle;

    // All the source buffers created by this object.
    readonly attribute SourceBufferList sourceBuffers;

    // Subset of sourceBuffers that provide data for the selected/enabled tracks.
    readonly attribute SourceBufferList activeSourceBuffers;

    readonly attribute ReadyState readyState;
    attribute unrestricted double duration;
    attribute EventHandler onsourceopen;
    attribute EventHandler onsourceended;
    attribute EventHandler onsourceclose;

    [Conditional=MEDIA_SOURCE_IN_WORKERS, CallWith=CurrentScriptExecutionContext]
    static readonly attribute boolean canConstructInDedicatedWorker;

    [NewObject] SourceBuffer addSourceBuffer(DOMString type);
    undefined removeSourceBuffer(SourceBuffer buffer);
    undefined endOfStream(optional EndOfStreamError error);
    undefined setLiveSeekableRange(double start, double end);
    undefined clearLiveSeekableRange();
    [CallWith=CurrentScriptExecutionContext] static boolean isTypeSupported(DOMString type);

};
